@@ -16,7 +16,7 @@ import sys
 import os
 import random
 import hashlib
-from jsonpath_ng import jsonpath, parse
+from jsonpath_ng import parse
 from typing import Optional, List
 from sqlalchemy.sql import and_, or_, not_
 
@@ -32,6 +32,14 @@ logger = logging.getLogger(__name__)
 coloredlogs.install(level=logging.INFO)
 
 
+def rand_sleep_time(val=2.0, diff=0.5):
+    return max(0.0001, val + random.uniform(0, 2*diff) - diff)
+
+
+def rand_sleep(val=2.0, diff=0.5):
+    time.sleep(rand_sleep_time(val, diff))
+
+
 def jsonpath(path, obj, allow_none=False):
     r = [m.value for m in parse(path).find(obj)]
     return r[0] if not allow_none else (r[0] if r else None)
@@ -64,6 +72,9 @@ class BoolJob:
     def is_halving(self):
         return '--halving' in self.cli
 
+    def __repr__(self):
+        return 'BoolJob(%s, %s, %s)' % (self.name, self.vinfo, self.idx)
+
 
 class BoolRes:
     def __init__(self, job, ret_code, js_res, is_halving, rejects=False, pval=None, alpha=None, stderr=None):
@@ -99,6 +110,11 @@ class BoolRunner:
         self.num_all_jobs = 0
 
     def init_config(self):
+        if not self.args.threads:
+            self.parallel_tasks = try_fnc(lambda: int(os.getenv('BOOLTEST_PARALLEL', None)))
+        if not self.parallel_tasks:
+            self.parallel_tasks = try_fnc(lambda: int(os.getenv('RTT_PARALLEL', None)))
+
         try:
             with open(self.args.rtt_config) as fh:
                 dt = fh.read()
@@ -109,11 +125,14 @@ class BoolRunner:
             if not self.bool_wrapper:
                 self.bool_wrapper = jsonpath("$.wrapper", self.bool_config, True)
 
-            if not self.args.threads:
-                self.parallel_tasks = try_fnc(lambda: int(os.getenv('RTT_PARALLEL', None)))
+            if not self.parallel_tasks:
+                self.parallel_tasks = jsonpath('$."toolkit-settings".execution."max-parallel-tests-booltest"', self.rtt_config, True)
             if not self.parallel_tasks:
                 self.parallel_tasks = jsonpath('$."toolkit-settings".execution."max-parallel-tests"', self.rtt_config, True) or 1
 
+            # TODO: add memory-per-file limit. From file size and BoolTest params we can estimate memory usage and limit
+            # number of processes. Works fine if input data is passed as a file (not stdin, not supported now anyway).
+
         except Exception as e:
             logger.error("Could not load RTT config %s at %s" % (e, self.args.rtt_config), exc_info=e)
 
@@ -223,6 +242,7 @@ class BoolRunner:
             return
 
         results = runner.out_acc
+        stderr = ("\n".join(runner.err_acc)).strip()
         buff = (''.join(results)).strip()
         try:
             js = json.loads(buff)
@@ -237,7 +257,7 @@ class BoolRunner:
                     logger.info('rejects: %s, at alpha %.5e' % (br.rejects, br.alpha))
                 except Exception as e:
                     logger.warning('BoolTest could not be evaluated, probably missing reference distribution: %s, %s'
-                                   % (job.name, e))
+                                   % (job, e))
                     br.alpha = 1.0
 
             else:
@@ -247,7 +267,9 @@ class BoolRunner:
             self.on_add_result(br)
 
         except Exception as e:
-            logger.error("Exception processing results: %s" % (e,), exc_info=e)
+            logger.error("Exception processing results: %s, job: %s, ret code: %s" % (e, job, runner.ret_code),
+                         exc_info=e)
+            logger.info("Failed BoolTest stderr: %s" % (stderr, ))
             logger.info("[[[%s]]]" % buff)
 
     def should_use_db(self):
@@ -267,16 +289,23 @@ class BoolRunner:
         self.res_cached.last_update = datetime.now()
         self.res_cached.all_jobs = self.num_all_jobs
         self.res_cached.done_jobs = len(self.results)
-        try:
-            if self.res_cached.id is None:
-                self.res_session.add(self.res_cached)
-            else:
-                self.res_session.merge(self.res_cached)
 
-            self.res_session.flush()
-            self.res_session.commit()
-        except Exception as e:
-            logger.warning("Could not update result cache: %s" % (e,), exc_info=e)
+        for i in range(10):
+            try:
+                if self.res_cached.id is None:
+                    self.res_session.add(self.res_cached)
+                else:
+                    self.res_session.merge(self.res_cached)
+
+                self.res_session.flush()
+                self.res_session.commit()
+                return  # <--- return on success
+            except Exception as e:
+                logger.warning("Could not update result cache: %s" % (e,), exc_info=e)
+                self.res_session.rollback()
+                rand_sleep(0.1, 0.05)
+
+        logger.warning("Cached result store failed")
 
     def on_results_ready(self):
         if not self.should_use_db():
@@ -377,9 +406,12 @@ class BoolRunner:
                     logger.info("Deleted cached results with ID %s" % (self.res_cached.id,))
             except Exception as e:
                 logger.warning("Unable to remove result cache entry: %s" % (e,), exc_info=e)
+                s.rollback()
 
         except Exception as e:
             logger.warning("Exception in storing results: %s" % (e,), exc_info=e)
+            s.rollback()
+            raise
 
         finally:
             silent_expunge_all(s)
@@ -448,7 +480,7 @@ class BoolRunner:
 
             jsres_obj = json.loads(self.res_cached.booltest_results)
             jsres = jsres_obj['results']
-            for jsbres in jsres:
+            for ix, jsbres in enumerate(jsres):
                 jsjob = try_fnc(lambda: jsbres["job"])
                 cjob = None
                 if jsjob:
@@ -466,7 +498,8 @@ class BoolRunner:
                                alpha=try_fnc(lambda: jsbres['alpha']),
                                stderr=try_fnc(lambda: jsbres['stderr']))
                 if not bres.js_res:
-                    logger.info("Loaded result failed, no js_res for id: %s" % (self.res_cached.id,))
+                    logger.info("Loaded result failed, no js_res for id: %s, rec: %s, job %s"
+                                % (self.res_cached.id, ix, cjob))
                     continue
 
                 self.results.append(bres)
@@ -538,7 +571,18 @@ class BoolRunner:
                 logger.info("Runner %s started, job queue size: %d, running: %s"
                             % (i, self.job_queue.qsize(), self.get_num_running()))
 
-        self.on_results_ready()
+        # attempt to store results several times
+        attempts = 30
+        for i in range(attempts):
+            try:
+                self.on_results_ready()
+                return
+            except Exception as e:
+                logger.warning("Could not store results, attempt: %s, err: %s" % (i, e), exc_info=e)
+                rand_sleep(2, 0.5)
+
+        # If here, results were not stored
+        raise ValueError("Could not store results")
 
     def main(self):
         logger.debug('App started')
